#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Render `NetworkX`_ graphs using `GraphViz`_"""

__author__ = "Timothy Shields"
__maintainer__ = "Timothy Shields"
__email__ = "Timothy.Shields@twosigma.com"
__version__ = "0.1.2"

from nxv import html_like, styles
from nxv._functional import chain, switch
from nxv._rendering import render
from nxv._style import Style, compose
from nxv._util import boundary, contrasting_color, neighborhood, to_ordered_graph


class GraphVizError(Exception):
    """Raised when a `GraphViz`_ run fails."""


class GraphVizInstallationNotFoundError(Exception):
    """Raised when a `GraphViz`_ installation is not found."""


class GraphVizAlgorithmNotFoundError(Exception):
    """Raised when a `GraphViz`_ algorithm is not found."""


__all__ = [
    "render",
    "Style",
    "compose",
    "chain",
    "switch",
    "neighborhood",
    "boundary",
    "to_ordered_graph",
    "contrasting_color",
    "styles",
    "html_like",
    "GraphVizInstallationNotFoundError",
    "GraphVizAlgorithmNotFoundError",
    "GraphVizError",
]



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Union

import networkx as nx

from nxv import _graphviz, _ipython
from nxv._functional import _apply
from nxv._style import Style, compose
from nxv._util import is_multi_graph
from nxv.html_like._html_like import HtmlLike, render_html_like

_root_style = Style(node=lambda u, d: {"label": str(u)})

COLOR_ATTRIBUTES = {
    "bgcolor",
    "color",
    "fillcolor",
    "fontcolor",
    "labelfontcolor",
    "pencolor",
}
LABEL_ATTRIBUTES = {"label", "headlabel", "taillabel"}


def _to_gv_string(value, attribute=None):
    if value is None:
        return '""'
    if attribute is not None:
        if attribute in COLOR_ATTRIBUTES and not isinstance(value, str):
            return _to_gv_string(color(value))
        if attribute in LABEL_ATTRIBUTES and isinstance(value, HtmlLike):
            return "<" + render_html_like(value) + ">"
    value = str(value)
    if not value:
        return '""'
    escape_char = "\\"
    need_escape = {'"', escape_char}

    # These characters break GraphViz
    broken_chars = {chr(0), chr(26)}

    if set(value) & (need_escape | broken_chars):
        value = "".join(
            escape_char + c if c in need_escape else c
            for c in value
            if c not in broken_chars
        )

    return f'"{value}"'


def _attributes_modifier(attrs):
    attrs_str = ", ".join(
        f"{k}={_to_gv_string(v, attribute=k)}" for k, v in attrs.items()
    )
    return f"[{attrs_str}]"


def _graph_identifier(graph_type, name):
    assert graph_type in {"graph", "digraph", "subgraph"}
    return f"{graph_type} {_to_gv_string(name)}"


def _graph_attrs_declaration(attrs):
    return f"graph {_attributes_modifier(attrs)};"


def _indent(value):
    return 4 * " " + value


def _block(prefix, lines):
    begin = (prefix + " {") if prefix else "{"
    end = "}"
    return [begin] + list(map(_indent, lines)) + [end]


def _default_subgraph_func(u, d):
    return None


def _to_gv(graph, style: Style, subgraph_func=None):
    """
    Serializes a `NetworkX`_ graph as a `GraphViz`_ string.

    :param graph: A `NetworkX`_ ``Graph``, ``DiGraph``, ``MultiGraph``, or ``MultiDiGraph``.
    :param style: A :class:`~nxv.Style` object.
    :param subgraph_func: An optional function ``f(u, d)`` that returns a subgraph key,
                          where ``u`` is a `NetworkX`_ node and ``d`` is its attribute dict.
    :return: The raw `GraphViz`_ string format to pass as input to one of the GraphViz layout algorithms.
    """

    if subgraph_func is None:
        subgraph_func = _default_subgraph_func

    graph_attrs = _apply(style.graph, graph, graph.graph)
    graph_type = graph_attrs.get(
        "type", "digraph" if nx.is_directed(graph) else "graph"
    )
    assert graph_type in {"graph", "digraph"}
    edge_str = {"graph": "--", "digraph": "->"}[graph_type]
    ids = {u: f"node{i:04}" for i, u in enumerate(graph.nodes())}

    no_subgraph_nodes = []
    subgraph_nodes = {}
    for u, d in graph.nodes(data=True):
        subgraph = _apply(subgraph_func, u, d)
        if subgraph is None:
            no_subgraph_nodes.append((u, d))
        else:
            subgraph_nodes.setdefault(subgraph, []).append((u, d))

    def node_declaration(u, d):
        node_attrs = _apply(style.node, u, d)
        return f"{ids[u]} {_attributes_modifier(node_attrs)};"

    def edge_declaration(*edge):
        u, v = edge[:2]
        edge_attrs = _apply(style.edge, *edge)
        return f"{ids[u]} {edge_str} {ids[v]} {_attributes_modifier(edge_attrs)};"

    def subgraph_declaration(subgraph, nodes, attrs):
        return _block(
            _graph_identifier("subgraph", attrs.get("name", str(subgraph))),
            [_graph_attrs_declaration(attrs)]
            + [node_declaration(u, d) for u, d in nodes],
        )

    edges_kwargs = (
        {"keys": True, "data": True} if is_multi_graph(graph) else {"data": True}
    )

    lines = _block(
        _graph_identifier(graph_type, graph_attrs.get("name", "G")),
        (
            [_graph_attrs_declaration(graph_attrs)]
            + [node_declaration(u, d) for u, d in no_subgraph_nodes]
            + [
                line
                for subgraph, nodes in subgraph_nodes.items()
                for line in subgraph_declaration(
                    subgraph, nodes, _apply(style.subgraph, subgraph)
                )
            ]
            + [edge_declaration(*edge) for edge in graph.edges(**edges_kwargs)]
        ),
    )

    return "\n".join(lines)


def render(
    graph: Union[nx.Graph, nx.DiGraph, nx.MultiGraph, nx.MultiDiGraph],
    style: Optional[Style] = None,
    *,
    algorithm: Optional[str] = None,
    format: Optional[str] = None,
    graphviz_bin: Optional[str] = None,
    subgraph_func=None,
) -> Optional[bytes]:
    """
    Render a `NetworkX`_ graph using `GraphViz`_.

    In a Jupyter notebook, this will automatically display as an SVG.

    :param graph: A `NetworkX`_ graph.
    :param style: A style specifying how graph nodes and edges should map to `GraphViz attributes`_.
    :param subgraph_func: An optional function ``f(u, d)`` that returns a subgraph key,
                          where ``u`` is a `NetworkX`_ node and ``d`` is its attribute dict.
                          If it returns ``None`` the node is not in any subgraph.
    :param algorithm: The `GraphViz`_ layout algorithm.
                      Valid options include
                      ``"circo"``, ``"dot"``, ``"fdp"``, ``"neato"``, ``"osage"``, ``"sfdp"``, ``"twopi"``.
                      Defaults to ``"dot"``.
    :param format: The `GraphViz`_ output format. Valid options include ``"svg"`` and ``"raw"``. In a Jupyter
                   notebook, prefixing the ``format`` with ``"ipython/"`` will automatically display the rendered
                   output.
                   When running in an interactive setting like a Jupyter notebook, the default is ``"ipython/svg"``.
                   Otherwise, this parameter is required.
    :param graphviz_bin: The ``bin`` directory of the `GraphViz`_ installation.
                         Defaults to the ``GRAPHVIZ_BIN`` environment variable.
                         If neither this parameter nor the ``GRAPHVIZ_BIN`` environment variable is set,
                         then nxv will try to autodetect the ``bin`` directory of the `GraphViz`_ installation.
                         This behavior is for convenience and should not be relied on in production settings.
    :return: If ``format`` is not an ``"ipython/*"`` format, the render output; otherwise, ``None``.
    :raises GraphVizInstallationNotFoundError: If nxv cannot find a `GraphViz`_ installation.
    :raises GraphVizAlgorithmNotFoundError: If nxv cannot find the specified algorithm in a `GraphViz`_ installation.
    :raises GraphVizError: If `GraphViz`_ failed to run on the given inputs.
    """
    if algorithm is None:
        algorithm = "dot"
    if not algorithm or not isinstance(algorithm, str):
        raise ValueError(
            "The algorithm parameter must be the str name of a valid GraphViz algorithm."
        )

    if format is None:
        if _ipython.is_execution_context():
            format = "ipython/svg"
        else:
            raise ValueError(
                "You must specify a format when not in an IPython execution context."
            )

    is_ipython_format = format.startswith("ipython/")
    if is_ipython_format:
        _ipython.assert_execution_context()
    graphviz_format = format.split("/", 1)[1] if is_ipython_format else format

    style = compose([_root_style, style])
    gv = _to_gv(graph, style, subgraph_func=subgraph_func)

    if graphviz_format == "raw":
        if is_ipython_format:
            print(gv)
            return None
        else:
            return gv

    output = _graphviz.run(gv, algorithm, graphviz_format, graphviz_bin)

    if is_ipython_format:
        _ipython.display(output, format)
        return None
    else:
        return output


def _clamp(a, b, value):
    return max(a, min(b, value))


def _to_byte(value):
    return _clamp(0, 255, int(256 * float(value)))


def color(channels):
    """
    Convert RGB or RGBA color channel values to a `GraphViz`_ color string.

    :param channels: The color channel values, either RGB or RGBA. Values should be in the range [0, 1].
    :return: The `GraphViz`_ color string.
    """
    assert len(channels) in (3, 4)
    return "#" + "".join(f"{_to_byte(value):02X}" for value in channels)



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
The ``nxv.html_like`` subpackage provides functions for building
`GraphViz HTML-like labels <https://graphviz.org/doc/info/shapes.html#html>`_.

The idiomatic import for this subpackage is:

.. code-block:: python

   import nxv.html_like as H

"""

from nxv.html_like._html_like import (
    bold,
    font,
    horizontal_rule,
    image,
    italic,
    join,
    line_break,
    overline,
    strikethrough,
    subscript,
    superscript,
    table,
    table_cell,
    table_row,
    underline,
    vertical_rule,
)

__all__ = [
    "join",
    "line_break",
    "font",
    "italic",
    "bold",
    "underline",
    "overline",
    "subscript",
    "superscript",
    "strikethrough",
    "table",
    "table_row",
    "horizontal_rule",
    "table_cell",
    "vertical_rule",
    "image",
]



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Iterable, Optional

from nxv._functional import chain


class Style:
    """
    A specification for how to style a `NetworkX`_ graph using `GraphViz`_.

    See the `GraphViz attributes`_ documentation for information on what attributes
    are available to use with the ``graph``, ``node``, ``edge``, and ``subgraph`` parameters.

    :param graph: An optional dict of `GraphViz graph attributes`_,
                  or a function ``f(g, d)`` that returns it,
                  in which ``g`` is the `NetworkX`_ graph and ``d`` is its attribute dict.
    :param node: An optional dict of `GraphViz node attributes`_,
                 or a function ``f(u, d)`` that returns it,
                 in which ``u`` is a `NetworkX`_ node and ``d`` is its attribute dict.
    :param edge: An optional dict of `GraphViz edge attributes`_,
                 or a function ``f(u, v, d)`` that returns it,
                 in which ``(u, v)`` is a `NetworkX`_ edge and ``d`` is its attribute dict.
                 If styling a graph with multi-edges,
                 the signature should be ``f(u, v, k, d)`` instead,
                 where ``k`` is the edge key.
    :param subgraph: An optional dict of `GraphViz subgraph attributes`_,
                     or a function ``f(s)`` that returns it,
                     in which ``s`` is a subgraph key.
                     This only applies when calling ``nxv.render`` with a ``subgraph_func``.
    """

    def __init__(self, *, graph=None, node=None, edge=None, subgraph=None):
        self.graph = graph or {}
        self.node = node or {}
        self.edge = edge or {}
        self.subgraph = subgraph or {}


def compose(styles: Iterable[Optional[Style]]) -> Style:
    """
    Compose a sequence of :class:`~nxv.Style` objects as a single :class:`~nxv.Style`.

    :param styles: An iterable of :class:`~nxv.Style` objects.
    :return: The composed :class:`~nxv.Style`.
    """
    styles = [style for style in styles if style]
    return Style(
        graph=chain(style.graph for style in styles),
        node=chain(style.node for style in styles),
        edge=chain(style.edge for style in styles),
        subgraph=chain(style.subgraph for style in styles),
    )



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import platform
import re
import shutil
from functools import lru_cache
from subprocess import PIPE, Popen
from typing import List, Optional


@lru_cache()
def is_windows() -> bool:
    return platform.system().lower() == "windows"


@lru_cache()
def get_windows_program_files() -> List[str]:
    possible = ["C:\\Program Files\\", "C:\\Program Files (x86)\\"]
    return [path for path in possible if os.path.isdir(path)]


@lru_cache()
def get_windows_graphviz_bins() -> List[str]:
    return [
        os.path.join(program_files, name, "bin")
        for program_files in get_windows_program_files()
        for name in os.listdir(program_files)
        if name.lower().startswith("graphviz")
        if os.path.isdir(os.path.join(program_files, name))
        if os.path.isdir(os.path.join(program_files, name, "bin"))
    ]


@lru_cache()
def try_get_graphviz_algorithm_path(graphviz_bin: str, algorithm: str) -> Optional[str]:
    for name in [algorithm, f"{algorithm}.exe"]:
        path = os.path.join(graphviz_bin, name)
        if os.path.isfile(path):
            return path
    return None


@lru_cache()
def get_graphviz_bins(graphviz_bin: Optional[str], algorithm: str) -> List[str]:
    from nxv import GraphVizInstallationNotFoundError

    # Always use an explicit value if specified
    if graphviz_bin:
        if not os.path.isdir(graphviz_bin):
            raise GraphVizInstallationNotFoundError(
                f"No GraphViz installation was found at the location specified "
                f"by the graphviz_bin parameter: {graphviz_bin}"
            )
        return [graphviz_bin]

    # Otherwise, always use an explicit environment variable if specified
    graphviz_bin = os.environ.get("GRAPHVIZ_BIN")
    if graphviz_bin:
        if not os.path.isdir(graphviz_bin):
            raise GraphVizInstallationNotFoundError(
                f"No GraphViz installation was found at the location specified "
                f"by the GRAPHVIZ_BIN environment variable: {graphviz_bin}"
            )
        return [graphviz_bin]

    # Otherwise, use directory containing the result of `which {algorithm}`
    graphviz_bin = shutil.which(algorithm)
    if graphviz_bin:
        graphviz_bin = os.path.dirname(graphviz_bin)
        return [graphviz_bin]

    # Otherwise, on Windows, look for the standard install locations
    if is_windows():
        graphviz_bins = get_windows_graphviz_bins()
        if not graphviz_bins:
            raise GraphVizInstallationNotFoundError(
                "No GraphViz installation was found at any of the standard Windows locations."
            )
        return graphviz_bins

    raise GraphVizInstallationNotFoundError(
        "No GraphViz installation was specified. "
        "Use either the graphviz_bin parameter or the GRAPHVIZ_BIN environment variable."
    )


@lru_cache()
def get_graphviz_algorithm_path(graphviz_bin: Optional[str], algorithm: str) -> str:
    from nxv import GraphVizAlgorithmNotFoundError

    graphviz_bins = get_graphviz_bins(graphviz_bin, algorithm)
    for graphviz_bin in graphviz_bins:
        path = try_get_graphviz_algorithm_path(graphviz_bin, algorithm)
        if path:
            return path
    graphviz_bins_str = " OR ".join(graphviz_bins)
    raise GraphVizAlgorithmNotFoundError(
        f"No GraphViz algorithm named {algorithm} was found in the GraphViz installation: {graphviz_bins_str}"
    )


def run(gv, algorithm, format, graphviz_bin):
    """
    Runs a `GraphViz`_ layout algorithm on a `GraphViz`_ string to product an output with the specified format.

    :param gv: A `GraphViz`_ string.
    :param algorithm: A `GraphViz`_ layout algorithm.
    :param format: A `GraphViz`_ output format.
    :param graphviz_bin: The bin directory of the `GraphViz`_ installation.
                         Defaults to the ``GRAPHVIZ_BIN`` environment variable.
    :return: The output bytes.
    :raises GraphVizError: If `GraphViz`_ failed to run on the given inputs.
    """
    from nxv import GraphVizError

    algorithm_path = get_graphviz_algorithm_path(graphviz_bin, algorithm)
    p = Popen(
        [algorithm_path, f"-T{format}"],
        stdin=PIPE,
        stdout=PIPE,
        stderr=PIPE,
    )
    stdin = gv.encode("utf-8")
    stdout, stderr = p.communicate(stdin)
    if p.returncode == 0:
        return stdout
    message = stderr.decode("utf-8")
    line_numbers = set(
        int(match.group(1)) for match in re.finditer(r"line (\d+)", message)
    )
    context_line_numbers = set(
        line_number + offset for line_number in line_numbers for offset in range(-2, 3)
    )
    if line_numbers:

        def prefix(line_number):
            return ">>>" if line_number in line_numbers else "   "

        message = "\n".join(
            [
                message,
                "   Line |",
                *(
                    f"{prefix(i)}{i:>4} | {line}"
                    for i, line in enumerate(gv.splitlines(), 1)
                    if i in context_line_numbers
                ),
            ]
        )
    raise GraphVizError(message)



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
def render_html_like_str(value):
    # It's very important we escape '&' first, as it is itself used in escaping.
    return (
        str(value)
        .replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
    )


def render_html_like(value):
    if isinstance(value, HtmlLike):
        return str(value)
    return render_html_like_str(value)


class HtmlLike:
    def __init__(self, name, children=None, attributes=None):
        self.name = name
        self.children = None if children is None else tuple(children)
        self.attributes = {k: v for k, v in (attributes or {}).items() if v is not None}

    def __str__(self):
        children = (
            None
            if self.children is None
            else "".join(render_html_like(child) for child in self.children)
        )
        attrs = "".join(
            f' {k}="{render_html_like_str(v)}"' for k, v in self.attributes.items()
        )
        if self.name is None:
            assert not attrs
            return children or ""
        if self.children is None:
            return f"<{self.name}{attrs}/>"
        return f"<{self.name}{attrs}>{children}</{self.name}>"

    def __repr__(self):
        return f"HtmlLike({self.name!r}, children={self.children!r}, attributes={self.attributes!r})"


def join(children):
    return HtmlLike(None, children=children)


def line_break(attributes=None):
    return HtmlLike("BR", attributes=attributes)


def font(content, attributes=None):
    return HtmlLike("FONT", children=[content], attributes=attributes)


def italic(content):
    return HtmlLike("I", children=[content])


def bold(content):
    return HtmlLike("B", children=[content])


def underline(content):
    return HtmlLike("U", children=[content])


def overline(content):
    return HtmlLike("O", children=[content])


def subscript(content):
    return HtmlLike("SUB", children=[content])


def superscript(content):
    return HtmlLike("SUP", children=[content])


def strikethrough(content):
    return HtmlLike("S", children=[content])


def table(rows, attributes=None):
    return HtmlLike("TABLE", children=rows, attributes=attributes)


def table_row(cells):
    return HtmlLike("TR", children=cells)


def horizontal_rule():
    return HtmlLike("HR")


def table_cell(content, attributes=None):
    return HtmlLike("TD", children=[content], attributes=attributes)


def vertical_rule():
    return HtmlLike("VR")


def image(attributes=None):
    return HtmlLike("IMG", attributes=attributes)



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utilities for interacting with the IPython execution context."""


def is_execution_context() -> bool:
    """Returns whether the current execution context is IPython."""
    try:
        return bool(__IPYTHON__)
    except NameError:
        return False


def assert_execution_context():
    """Asserts that the current execution context is IPython."""
    if not is_execution_context():
        raise ValueError(
            "Must be running in an IPython execution context "
            "to use an ipython/* format."
        )


def display(data, format):
    """Display SVG or image data in the current IPython execution context."""
    assert_execution_context()
    from IPython.display import SVG, Image, display

    formats = {"ipython/svg": SVG, "ipython/png": Image}
    constructor = formats.get(format, None)
    if constructor is None:
        raise ValueError(
            f"Invalid IPython format {format!r}. "
            f"Valid IPython formats are {set(formats)}."
        )
    display(constructor(data))



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from nxv._style import Style


def _verbose_edge(*args):
    if len(args) == 3:
        u, v, d = args
        return {"label": str(d)}
    else:
        u, v, e, d = args
        return {"label": f"{e}\n{d}"}


def verbose() -> Style:
    """
    Get a verbose :class:`~nxv.Style` that shows all of the data in a graph.

    :return: A verbose :class:`~nxv.Style`.
    """
    return Style(
        graph=lambda g, d: {"label": f"Graph Attributes: {d}", "labelloc": "t"},
        node=lambda u, d: {"label": f"{u}\n{d}", "shape": "box"},
        edge=_verbose_edge,
    )



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Union

from nxv._style import Style


def font(
    fontname: Optional[str] = None, fontsize: Optional[Union[int, float]] = None
) -> Style:
    """
    Styles text in a graph using the given font.

    :param fontname: Optional font name.
    :param fontsize: Optional font size.
    :return: A :class:`~nxv.Style` object that applies this font.
    """
    d = {}
    if fontname is not None:
        d["fontname"] = fontname
    if fontsize is not None:
        d["fontsize"] = fontsize
    return Style(graph=d, node=d, edge=d)



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utilities for composing functions."""


def _apply(func, *args, **kwargs):
    """
    Apply a maybe-function to positional and named arguments.

    The result is ``func(*args, **kwargs)`` if ``callable(func)``, otherwise ``func``.

    :param func: A value that is optionally callable.
    :param args: The positional arguments passed to ``func``, if it is callable.
    :param kwargs: The named arguments passed to ``func``, if it is callable.
    :return: ``func(*args, **kwargs)`` if ``callable(func)``. Otherwise, returns ``func``.
    """
    if callable(func):
        return func(*args, **kwargs)
    else:
        return func


def chain(funcs):
    """
    Chain a sequence of dict-returning functions together to form a new dict-returning function.

    The result is a function ``f(*args, **kwargs)`` that returns
    ``{**apply(funcs[0], *args, **kwargs), **apply(funcs[1], *args, **kwargs), ...}``.

    :param funcs: An iterable of functions that return dicts.
    :return: A function ``f(*args, **kwargs)`` that returns
             ``{**apply(funcs[0], *args, **kwargs), **apply(funcs[1], *args, **kwargs), ...}``.
    """
    funcs = list(funcs)

    def func(*args, **kwargs):
        result = {}
        for f in funcs:
            result.update(_apply(f, *args, **kwargs))
        return result

    return func


def switch(key, funcs, *, default=None):
    """
    Combine a dict of keyed functions to form a new function.

    The result is a function ``f(*args, **kwargs)`` that returns
    ``apply(funcs[key(*args, **kwargs)], *args, **kwargs)``.

    If ``key(*args, **kwargs)`` is not in ``funcs`` but ``default`` is present,
    ``apply(default, *args, **kwargs)`` will be returned instead.

    :param key: The key selector function.
    :param funcs: The mapping from keys to functions.
    :param default: An optional default function for keys that do not appear in ``funcs``.
    :return: The function ``f(*args, **kwargs)`` that returns ``apply(funcs[key(*args, **kwargs)], *args, **kwargs)``.
    """

    def func(*args, **kwargs):
        k = key(*args, **kwargs)
        if k not in funcs and default is not None:
            return _apply(default, *args, **kwargs)
        return _apply(funcs[k], *args, **kwargs)

    return func



#
# Copyright 2020 Two Sigma Open Source, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import heapq
from collections import OrderedDict

import networkx as nx


def _default_cost(u, v):
    return 1


def uniform_cost_traversal(sources, traverse, *, cost=None):
    """
    A Uniform-Cost Search without a goal.

    :param sources: An iterable of source nodes.
    :param traverse: A function ``f(u)`` that returns the successors of ``u``.
    :param cost: A function ``f(u, v)`` specifying the cost of traversing from ``u`` to ``v``.
    :return: Generates ``(c, u)`` where ``u`` is a traversed node and ``c`` is the cost of reaching that node.
    """
    if cost is None:
        cost = _default_cost

    visited = set()
    heap = []
    index = [0]

    def push(cu, u):
        heapq.heappush(heap, (cu, index[0], u))
        index[0] += 1

    def pop():
        cu, _, u = heapq.heappop(heap)
        return cu, u

    for u in sources:
        push(0, u)
    while heap:
        cu, u = pop()
        if u in visited:
            continue
        yield cu, u
        visited.add(u)
        for v in traverse(u):
            cv = cu + cost(u, v)
            push(cv, v)


def neighborhood(graph, nodes, *, radius=None, cost=None):
    """
    Get the subgraph in the neighborhood of the specified nodes.

    This is useful for viewing a small portion of a large graph.

    :param graph: A graph.
    :param nodes: An iterable of nodes.
    :param radius: The size of the neighborhood.
    :param cost: A function ``f(u, v)`` specifying the cost of traversing from ``u`` to ``v``.
    :return: The neighborhood subgraph.
    """

    def traverse(u):
        return nx.all_neighbors(graph, u)

    def subgraph_nodes():
        for cu, u in uniform_cost_traversal(nodes, traverse, cost=cost):
            if radius is not None and cu > radius:
                return
            yield u

    return graph.subgraph(subgraph_nodes())


def boundary(graph, subgraph):
    """
    Get the nodes in the subgraph that have neighbors in the graph but not in the subgraph.

    This is useful for conditionally styling nodes at the boundary of a subgraph. For example:

    ::

        boundary = nxv.boundary(graph, subgraph)
        style = nxv.Style(node=lambda u, d: {
            'style': 'dashed' if u in boundary else 'solid',
        })
        nxv.render(subgraph, style)


    :param graph: A graph.
    :param subgraph: A subgraph of the graph.
    :return: The nodes in the subgraph that have neighbors in the graph but not in the subgraph.
    """
    if not all(graph.has_node(node) for node in subgraph.nodes()):
        if all(subgraph.has_node(node) for node in graph.nodes()):
            raise ValueError(
                "The 'graph' argument is a proper subgraph of the 'subgraph' argument. This is likely "
                "because the arguments to boundary were passed in the wrong order."
            )
        else:
            raise ValueError("The subgraph contains nodes not in the graph.")
    return {
        node for node in subgraph.nodes() if subgraph.degree(node) < graph.degree(node)
    }


# directed, multi, ordered
GRAPH_TYPES = {
    (False, False, False): nx.Graph,
    (True, False, False): nx.DiGraph,
    (False, True, False): nx.MultiGraph,
    (True, True, False): nx.MultiDiGraph,
    (False, False, True): nx.OrderedGraph,
    (True, False, True): nx.OrderedDiGraph,
    (False, True, True): nx.OrderedMultiGraph,
    (True, True, True): nx.OrderedMultiDiGraph,
}
MULTI_GRAPH_TYPES = (
    nx.MultiGraph,
    nx.MultiDiGraph,
    nx.OrderedMultiGraph,
    nx.OrderedMultiDiGraph,
)
ORDERED_GRAPH_TYPES = (
    nx.OrderedGraph,
    nx.OrderedDiGraph,
    nx.OrderedMultiGraph,
    nx.OrderedMultiDiGraph,
)


def is_directed_graph(graph):
    """
    Get whether the graph has directed edges.

    :param graph: The graph.
    :return: True if the graph has directed edges.
    """
    return nx.is_directed(graph)


def is_multi_graph(graph):
    """
    Get whether the graph has multi-edges.

    :param graph: The graph.
    :return: True if the graph has multi-edges.
    """
    return isinstance(graph, MULTI_GRAPH_TYPES)


def is_ordered_graph(graph):
    """
    Get whether the graph has ordered nodes and edges.

    :param graph: The graph.
    :return: True if the graph has ordered nodes and edges.
    """
    return isinstance(graph, ORDERED_GRAPH_TYPES)


def create_graph(directed, multi, ordered):
    """
    Create an empty graph of the specified kind.

    :param directed: Should the graph have directed edges?
    :param multi: Should the graph allow multi-edges?
    :param ordered: Should the nodes and edges of the graph be ordered?
    :return: An empty graph of the specified type.
    """
    return GRAPH_TYPES[directed, multi, ordered]()


def to_ordered_graph(graph, node_key=None, edge_key=None, attr_key=None):
    """
    Create an ordered copy of the specified graph, with nodes and edges ordered by the specified key functions.

    :param graph: The graph to order.
    :param node_key: The node key function, ``node_key(u, d)``. Defaults to the identity function.
    :param edge_key: The edge key function, ``edge_key(u, v, d)``. If the graph has multi-edges, the signature should be
                     ``edge_key(u, v, k, d)`` instead, where ``k`` is the edge key. Defaults to the identity function.
    :param attr_key: The attribute key function, ``attr_key(k, v)``. Defaults to the identity function.
    :return: A copy of the graph with the nodes and edges ordered by the specified key functions.
    """
    result = create_graph(is_directed_graph(graph), is_multi_graph(graph), True)
    result.graph.update(graph.graph)
    for u, d in sorted(graph.nodes(data=True), key=node_key):
        result.add_node(u)
        result.nodes[u].update(OrderedDict(sorted(d.items(), key=attr_key)))
    if is_multi_graph(graph):
        for u, v, k, d in sorted(graph.edges(keys=True, data=True), key=edge_key):
            result.add_edge(u, v, k)
            result.edges[u, v, k].update(OrderedDict(sorted(d.items(), key=attr_key)))
    else:
        for u, v, d in sorted(graph.edges(data=True), key=edge_key):
            result.add_edge(u, v)
            result.edges[u, v].update(OrderedDict(sorted(d.items(), key=attr_key)))
    return result


def _relative_luminance_helper(x):
    x = max(0, min(x, 1))
    if x <= 0.03928:
        return x / 12.92
    return ((x + 0.055) / 1.055) ** 2.4


def _relative_luminance(channels):
    # See: https://www.w3.org/TR/2008/REC-WCAG20-20081211/#relativeluminancedef
    weights = (0.2126, 0.7152, 0.0722)
    return sum(
        weight * _relative_luminance_helper(channel)
        for weight, channel in zip(weights, channels)
    )


def _contrast_ratio(l1, l2):
    # See: https://www.w3.org/TR/2008/REC-WCAG20-20081211/#contrast-ratiodef
    if l1 < l2:
        l1, l2 = l2, l1
    return (l1 + 0.05) / (l2 + 0.05)


def contrasting_color(channels, *, options=None):
    """
    Get a color that most contrasts with a specified color.

    :param channels: The RGB or RGBA color channels. Values should be in the range [0, 1].
    :param options: The possible contrasting colors. Defaults to black and white.
    :return: The color option that most contrasts the input color.
    """
    if options is None:
        options = ((0, 0, 0), (1, 1, 1))
    if not options:
        return (0, 0, 0)
    lum = _relative_luminance(channels)
    return max(
        options, key=lambda option: _contrast_ratio(lum, _relative_luminance(option))
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nxv', 'nxv.html_like', 'nxv.styles']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.5,<3.0']

setup_kwargs = {
    'name': 'nxv',
    'version': '0.1.2',
    'description': 'Render NetworkX graphs using GraphViz',
    'long_description': '<img src="./docs/_static/logo/logo.svg" align="right">\n\n# nxv\n\n[![PyPI Status](https://img.shields.io/pypi/v/nxv.svg)](https://pypi.python.org/pypi/nxv)\n![Tests](https://github.com/twosigma/nxv/workflows/Tests/badge.svg)\n[![Documentation Status](https://readthedocs.org/projects/nxv/badge/?version=latest)](https://nxv.readthedocs.io/en/latest/?badge=latest)\n[![Codecov](https://codecov.io/gh/twosigma/nxv/branch/main/graph/badge.svg)](https://codecov.io/gh/twosigma/nxv)\n\n\nRender NetworkX graphs using GraphViz.\n\n# Documentation\n\nhttps://nxv.readthedocs.io/\n\n# Basic Usage\n\n    import networkx as nx\n    import nxv\n    \n    graph = nx.Graph()\n    graph.add_edge("A", "B")\n    graph.add_edge("B", "C")\n    graph.add_edge("C", "D")\n    graph.add_edge("B", "E")\n\n    style = nxv.Style(\n        graph={"rankdir": "LR"},\n        node=lambda u, d: {"shape": "circle" if u in "AEIOU" else "square"},\n        edge=lambda u, v, d: {"style": "dashed", "label": u + v},\n    )\n    \n    nxv.render(graph, style)\n\n<img src="./docs/_static/example/quickstart_graph_functional_style.svg">\n\n# Installation\n\n    pip install nxv\n\n# Development\n\nThis repository uses\n[Poetry](https://python-poetry.org/) and\n[Nox](https://nox.thea.codes/en/stable/)\nto manage the development environment and builds.\n\nTo list all Nox sessions:\n\n    python -m nox --list-sessions\n\nTo run the black code formatter:\n\n    python -m nox -rs black\n\nTo lint using flake8:\n\n    python -m nox -rs lint\n\nTo run the test suite:\n\n    python -m nox -rs tests\n\nTo build the documentation:\n\n    python -m nox -rs docs\n',
    'author': 'Timothy Shields',
    'author_email': 'Timothy.Shields@twosigma.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/twosigma/nxv',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
